import shutil
import os
import glob
import subprocess as sp

import numpy as np
import stingray
from astropy import log
import pytest
from stingray.events import EventList
import hendrics as hen
from hendrics.tests import _dummy_par
from hendrics import (
    calibrate,
    colors,
    io,
    lcurve,
    modeling,
    plot,
    read_events,
    rebin,
    save_as_xspec,
    sum_fspec,
    timelags,
    varenergy,
)

from hendrics.base import touch
from hendrics.fspec import calc_cpds, calc_pds

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

HEN_FILE_EXTENSION = hen.io.HEN_FILE_EXTENSION

log.setLevel("DEBUG")
# log.basicConfig(filename='HEN.log', level=log.DEBUG, filemode='w')


def test_pds_fails_noclobber_exists():
    touch("bububu")
    with pytest.warns(UserWarning) as record:
        calc_pds("bla.p", 512, outname="bububu", noclobber=True)
    assert np.any(
        ["File exists, and noclobber" in r.message.args[0] for r in record]
    )
    os.unlink("bububu")


def test_cpds_fails_noclobber_exists():
    touch("bububu")
    with pytest.warns(UserWarning) as record:
        calc_cpds("bla.p", "blu.p", 512, outname="bububu", noclobber=True)
    assert np.any(
        ["File exists, and noclobber" in r.message.args[0] for r in record]
    )
    os.unlink("bububu")


def test_distributed_pds():
    events = EventList(
        np.sort(np.random.uniform(0, 1000, 1000)), gti=np.asarray([[0, 1000]])
    )
    single_periodogram = stingray.AveragedPowerspectrum(
        events, segment_size=100, dt=0.1, norm="leahy"
    )
    pds_iterable = hen.fspec._provide_periodograms(events, 100, 0.1, "leahy")
    pds_distr = hen.fspec.average_periodograms(pds_iterable)
    assert np.allclose(pds_distr.power, single_periodogram.power)
    assert np.allclose(pds_distr.power_err, single_periodogram.power_err)
    assert np.allclose(pds_distr.freq, single_periodogram.freq)
    assert pds_distr.m == single_periodogram.m


def test_distributed_cpds():
    events1 = EventList(
        np.sort(np.random.uniform(0, 1000, 1000)), gti=np.asarray([[0, 1000]])
    )
    events2 = EventList(
        np.sort(np.random.uniform(0, 1000, 1000)), gti=np.asarray([[0, 1000]])
    )
    single_periodogram = stingray.AveragedCrossspectrum(
        events1, events2, segment_size=100, dt=0.1, norm="leahy"
    )
    pds_iterable = hen.fspec._provide_cross_periodograms(
        events1, events2, 100, 0.1, "leahy"
    )
    pds_distr = hen.fspec.average_periodograms(pds_iterable)
    assert np.allclose(pds_distr.power, single_periodogram.power)
    assert np.allclose(pds_distr.power_err, single_periodogram.power_err)
    assert np.allclose(pds_distr.freq, single_periodogram.freq)
    assert pds_distr.m == single_periodogram.m


class TestFullRun(object):
    """Test how command lines work.

    Usually considered bad practice, but in this
    case I need to test the full run of the codes, and files depend on each
    other.
    Inspired by http://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order

    When command line is missing, uses some function calls
    """  # NOQA

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")

        cls.par = _dummy_par("bubububu.par")

        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileB = os.path.join(
            cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileAcal = os.path.join(
            cls.datadir,
            "monol_testA_nustar_fpma_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.ev_fileBcal = os.path.join(
            cls.datadir,
            "monol_testB_nustar_fpmb_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.par = _dummy_par("bubububu.par")
        command = "{0} {1}".format(
            os.path.join(cls.datadir, "monol_testA.evt"),
            os.path.join(cls.datadir, "monol_testB.evt"),
        )
        hen.read_events.main(command.split())
        command = "{} {} -r {}".format(
            os.path.join(
                cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(
                cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(cls.datadir, "test.rmf"),
        )
        hen.calibrate.main(command.split())
        cls.lcA = os.path.join(
            os.path.join(
                cls.datadir, "monol_testA_nustar_fpma_lc" + HEN_FILE_EXTENSION
            )
        )
        cls.lcB = os.path.join(
            os.path.join(
                cls.datadir, "monol_testB_nustar_fpmb_lc" + HEN_FILE_EXTENSION
            )
        )
        cls.lcAfilt = os.path.join(
            os.path.join(
                cls.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )
        cls.lcBfilt = os.path.join(
            os.path.join(
                cls.datadir, "monol_testB_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )
        command = (
            "{} -e 3 50 --safe-interval 100 300  --nproc 2 -b 0.5 " "-o {}"
        ).format(cls.ev_fileAcal, cls.lcAfilt)
        hen.lcurve.main(command.split())
        command = (
            "{} -e 3 50 --safe-interval 100 300  --nproc 2 -b 0.5 " "-o {}"
        ).format(cls.ev_fileBcal, cls.lcBfilt)
        hen.lcurve.main(command.split())
        command = ("{} --nproc 2 -b 0.5 " "-o {}").format(
            cls.ev_fileAcal, cls.lcA
        )
        hen.lcurve.main(command.split())
        command = ("{}  --nproc 2 -b 0.5 " "-o {}").format(
            cls.ev_fileBcal, cls.lcB
        )
        hen.lcurve.main(command.split())

        cls.pdsA = os.path.join(
            cls.datadir, "monol_testA_E3-50_pds" + HEN_FILE_EXTENSION
        )
        cls.pdsB = os.path.join(
            cls.datadir, "monol_testB_E3-50_pds" + HEN_FILE_EXTENSION
        )
        cls.cpds = os.path.join(
            cls.datadir, "monol_test_E3-50_cpds" + HEN_FILE_EXTENSION
        )

        command = "{} {} -f 128 -k PDS --save-all --norm leahy".format(
            cls.lcAfilt, cls.lcBfilt
        )
        hen.fspec.main(command.split())

        command = "{} {} -f 128 -k CPDS --save-all --norm leahy".format(
            cls.lcAfilt, cls.lcBfilt
        )
        hen.fspec.main(command.split())
        assert os.path.exists(cls.cpds)
        assert os.path.exists(cls.pdsA)
        assert os.path.exists(cls.pdsB)

    def test_pds_leahy_dtbig(self):
        """Test PDS production."""
        lc = self.lcAfilt
        hen.io.main([lc])
        command = "{0} -f 128 -k PDS --save-all --norm leahy -b {1}".format(
            lc, 1
        )
        hen.fspec.main(command.split())

        assert os.path.exists(
            os.path.join(
                self.datadir, "monol_testA_E3-50_pds" + HEN_FILE_EXTENSION
            )
        )

    def test_pds_leahy(self):
        """Test PDS production."""
        evdata = self.ev_fileA
        lcdata = self.lcA

        command = "{0} -f 128 -k PDS --norm leahy -b 0.5".format(evdata)
        hen.fspec.main(command.split())
        command = "{0} -f 128 -k PDS --save-all --norm leahy".format(lcdata)
        hen.fspec.main(command.split())

        evout = evdata.replace("_ev", "_pds")
        lcout = lcdata.replace("_lc", "_pds")
        assert os.path.exists(evout)
        assert os.path.exists(lcout)
        evpds = hen.io.load_pds(evout)
        lcpds = hen.io.load_pds(lcout)
        assert np.allclose(evpds.power, lcpds.power)

    @pytest.mark.parametrize("data_kind", ["events", "lc"])
    def test_pds(self, data_kind):
        """Test PDS production."""
        if data_kind == "events":
            labelA = "nustar_fpma_ev"
            labelB = "nustar_fpmb_ev"
        else:
            labelA = labelB = "E3-50_lc"
        command = (
            "{0} {1} -f 128 --save-all --save-dyn -k PDS "
            "--norm frac".format(
                os.path.join(self.datadir, f"monol_testA_{labelA}")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, f"monol_testB_{labelB}")
                + HEN_FILE_EXTENSION,
            )
        )
        hen.fspec.main(command.split())

        out_labelA = labelA.replace("_ev", "_pds").replace("_lc", "_pds")
        out_labelB = labelB.replace("_ev", "_pds").replace("_lc", "_pds")
        assert os.path.exists(
            os.path.join(
                self.datadir, f"monol_testB_{out_labelB}" + HEN_FILE_EXTENSION
            )
        )
        assert os.path.exists(
            os.path.join(self.datadir, f"monol_testA_{out_labelA}")
            + HEN_FILE_EXTENSION
        )

    @pytest.mark.parametrize("kind", ["PDS", "CPDS"])
    def test_pds_events_big(self, kind):
        """Test PDS production."""
        labelA = "nustar_fpma_ev"
        labelB = "nustar_fpmb_ev"

        command = (
            "{0} {1} -f 128 --save-all --save-dyn -k {2} "
            "--norm frac --test".format(
                os.path.join(self.datadir, f"monol_testA_{labelA}")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, f"monol_testB_{labelB}")
                + HEN_FILE_EXTENSION,
                kind,
            )
        )
        hen.fspec.main(command.split())

    def test_cpds_ignore_instr(self):
        """Test CPDS production."""
        command = (
            "{0} {1} -f 128 --save-dyn -k CPDS,lag --save-all --ignore-instr"
            " -o {2} --debug".format(
                os.path.join(self.datadir, "monol_testA_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_testB_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "ignore_instr")
                + HEN_FILE_EXTENSION,
            )
        )

        hen.fspec.main(command.split())

    def test_cpds_rms_norm(self):
        """Test CPDS production."""
        command = (
            "{0} {1} -f 128 --save-dyn -k CPDS --save-all "
            "--norm rms -o {2}".format(
                os.path.join(self.datadir, "monol_testA_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_testB_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_test_E3-50"),
            )
        )

        hen.fspec.main(command.split())

    def test_cpds_wrong_norm(self):
        """Test CPDS production."""
        command = (
            "{0} {1} -f 128 --save-dyn -k CPDS --norm blablabla -o {2}".format(
                os.path.join(self.datadir, "monol_testA_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_testB_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_test_E3-50"),
            )
        )
        with pytest.warns(UserWarning) as record:
            hen.fspec.main(command.split())

        assert np.any(
            [
                "Beware! Unknown normalization" in r.message.args[0]
                for r in record
            ]
        )

    def test_cpds_dtbig(self):
        """Test CPDS production."""
        command = (
            "{0} {1} -f 128 --save-dyn -k CPDS --save-all --norm "
            "frac -o {2}".format(
                os.path.join(self.datadir, "monol_testA_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_testB_E3-50_lc")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_test_E3-50"),
            )
        )
        command += " -b 1"
        hen.fspec.main(command.split())

    def test_dumpdynpds(self):
        """Test dump dynamical PDSs."""
        command = (
            "--noplot "
            + os.path.join(self.datadir, "monol_testA_E3-50_pds")
            + HEN_FILE_EXTENSION
        )
        with pytest.raises(NotImplementedError):
            hen.fspec.dumpdyn_main(command.split())

    def test_sumpds(self):
        """Test the sum of pdss."""
        hen.sum_fspec.main(
            [
                os.path.join(self.datadir, "monol_testA_E3-50_pds")
                + HEN_FILE_EXTENSION,
                os.path.join(self.datadir, "monol_testB_E3-50_pds")
                + HEN_FILE_EXTENSION,
                "-o",
                os.path.join(
                    self.datadir, "monol_test_sum" + HEN_FILE_EXTENSION
                ),
            ]
        )

    def test_dumpdyncpds(self):
        """Test dump dynamical PDSs."""
        command = (
            "--noplot "
            + os.path.join(self.datadir, "monol_test_E3-50_cpds")
            + HEN_FILE_EXTENSION
        )
        with pytest.raises(NotImplementedError):
            hen.fspec.dumpdyn_main(command.split())

    def test_rebinpds(self):
        """Test PDS rebinning 1."""
        command = "{0} -r 2".format(
            os.path.join(self.datadir, "monol_testA_E3-50_pds")
            + HEN_FILE_EXTENSION
        )
        hen.rebin.main(command.split())
        os.path.exists(
            os.path.join(
                self.datadir,
                "monol_testA_E3-50_pds_rebin2" + HEN_FILE_EXTENSION,
            )
        )

    def test_rebinpds_geom(self):
        """Test geometrical PDS rebinning."""
        command = "{0} {1} -r 1.03".format(
            os.path.join(self.datadir, "monol_testA_E3-50_pds")
            + HEN_FILE_EXTENSION,
            os.path.join(self.datadir, "monol_testB_E3-50_pds")
            + HEN_FILE_EXTENSION,
        )
        hen.rebin.main(command.split())
        os.path.exists(
            os.path.join(
                self.datadir,
                "monol_testA_E3-50_pds_rebin1.03" + HEN_FILE_EXTENSION,
            )
        )
        os.path.exists(
            os.path.join(
                self.datadir,
                "monol_testB_E3-50_pds_rebin1.03" + HEN_FILE_EXTENSION,
            )
        )

    def test_rebincpds(self):
        """Test CPDS rebinning."""
        command = "{0} -r 2".format(
            os.path.join(self.datadir, "monol_test_E3-50_cpds")
            + HEN_FILE_EXTENSION
        )
        hen.rebin.main(command.split())
        os.path.exists(
            os.path.join(
                self.datadir,
                "monol_test_E3-50_cpds_rebin2" + HEN_FILE_EXTENSION,
            )
        )

    def test_rebincpds_geom(self):
        """Test CPDS geometrical rebinning."""
        command = "{0} -r 1.03".format(
            os.path.join(self.datadir, "monol_test_E3-50_cpds")
            + HEN_FILE_EXTENSION
        )
        hen.rebin.main(command.split())
        os.path.exists(
            os.path.join(
                self.datadir,
                "monol_test_E3-50_cpds_rebin1.03" + HEN_FILE_EXTENSION,
            )
        )

    def test_save_lags(self):
        fname = os.path.join(
            self.datadir, "monol_test_E3-50_cpds_rebin2" + HEN_FILE_EXTENSION
        )
        hen.timelags.main([fname])
        out = hen.base.hen_root(fname) + "_lags.qdp"
        os.path.exists(out)

    def test_fit_pds(self):
        modelstring = """
from astropy.modeling import models
model = models.Const1D()
        """
        modelfile = "bubu__model__.py"
        with open(modelfile, "w") as fobj:
            print(modelstring, file=fobj)
        pdsfile1 = os.path.join(
            self.datadir, "monol_testA_E3-50_pds" + HEN_FILE_EXTENSION
        )
        pdsfile2 = os.path.join(
            self.datadir, "monol_testB_E3-50_pds" + HEN_FILE_EXTENSION
        )

        command = "{0} {1} -m {2} --frequency-interval 0 10".format(
            pdsfile1, pdsfile2, modelfile
        )
        hen.modeling.main_model(command.split())

        out0 = os.path.join(self.datadir, "monol_testA_E3-50_pds_bestfit.p")
        out1 = os.path.join(self.datadir, "monol_testB_E3-50_pds_bestfit.p")
        assert os.path.exists(out0)
        assert os.path.exists(out1)
        m, k, c = hen.io.load_model(
            os.path.join(self.datadir, "monol_testB_E3-50_pds_bestfit.p")
        )
        assert hasattr(m, "amplitude")
        os.unlink(out0)
        os.unlink(out1)

        out0 = os.path.join(
            self.datadir, "monol_testA_E3-50_pds_fit" + HEN_FILE_EXTENSION
        )
        out1 = os.path.join(
            self.datadir, "monol_testB_E3-50_pds_fit" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(out0)
        assert os.path.exists(out1)
        spec = hen.io.load_pds(out0)
        assert hasattr(spec, "best_fits")

    def test_fit_cpds(self):
        modelstring = """
from astropy.modeling import models
model = models.Const1D()
        """
        modelfile = "bubu__model__.py"
        with open(modelfile, "w") as fobj:
            print(modelstring, file=fobj)
        pdsfile1 = os.path.join(
            self.datadir, "monol_test_E3-50_cpds" + HEN_FILE_EXTENSION
        )

        command = "{0} -m {1} --frequency-interval 0 10".format(
            pdsfile1, modelfile
        )
        hen.modeling.main_model(command.split())

        out0 = os.path.join(self.datadir, "monol_test_E3-50_cpds_bestfit.p")
        assert os.path.exists(out0)
        m, k, c = hen.io.load_model(out0)
        assert hasattr(m, "amplitude")
        os.unlink(out0)

        out0 = os.path.join(
            self.datadir, "monol_test_E3-50_cpds_fit" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(out0)
        spec = hen.io.load_pds(out0)
        assert hasattr(spec, "best_fits")

    def test_fit_pds_f_no_of_intervals_invalid(self):
        modelstring = """
from astropy.modeling import models
model = models.Const1D()
        """
        modelfile = "bubu__model__.py"
        with open(modelfile, "w") as fobj:
            print(modelstring, file=fobj)
        pdsfile1 = os.path.join(
            self.datadir, "monol_testA_E3-50_pds" + HEN_FILE_EXTENSION
        )
        pdsfile2 = os.path.join(
            self.datadir, "monol_testB_E3-50_pds" + HEN_FILE_EXTENSION
        )

        command = "{0} {1} -m {2} --frequency-interval 0 1 9".format(
            pdsfile1, pdsfile2, modelfile
        )
        with pytest.raises(ValueError) as excinfo:
            hen.modeling.main_model(command.split())

        assert "Invalid number of frequencies specified" in str(excinfo.value)

    def test_savexspec(self):
        """Test save as Xspec 1."""
        command = "{0}".format(
            os.path.join(self.datadir, "monol_testA_E3-50_pds_rebin2")
            + HEN_FILE_EXTENSION
        )
        hen.save_as_xspec.main(command.split())
        os.path.exists(
            os.path.join(self.datadir, "monol_testA_E3-50_pds_rebin2.pha")
        )

    def test_savexspec_geom(self):
        """Test save as Xspec 2."""
        command = "{0}".format(
            os.path.join(self.datadir, "monol_test_E3-50_cpds_rebin1.03")
            + HEN_FILE_EXTENSION
        )
        hen.save_as_xspec.main(command.split())

        os.path.exists(
            os.path.join(self.datadir, "monol_test_E3-50_cpds_rebin1.03.pha")
        )
        os.path.exists(
            os.path.join(
                self.datadir, "monol_test_E3-50_cpds_rebin1.03_lags.pha"
            )
        )

    def test_plot_lin(self):
        """Test plotting with linear axes."""
        pname = (
            os.path.join(self.datadir, "monol_testA_E3-50_pds_fit")
            + HEN_FILE_EXTENSION
        )
        cname = (
            os.path.join(self.datadir, "monol_test_E3-50_cpds_fit")
            + HEN_FILE_EXTENSION
        )
        lname = (
            os.path.join(self.datadir, "monol_testA_E3-50_lc")
            + HEN_FILE_EXTENSION
        )
        hen.plot.main(
            [
                pname,
                cname,
                lname,
                "--noplot",
                "--xlin",
                "--ylin",
                "-o",
                "dummy.qdp",
            ]
        )
        hen.plot.main(
            [
                lname,
                "--noplot",
                "--axes",
                "time",
                "counts",
                "--xlin",
                "--ylin",
                "-o",
                "dummy.qdp",
            ]
        )

    def test_plot_log(self):
        """Test plotting with log axes."""
        pname = os.path.join(
            self.datadir,
            "monol_testA_E3-50_pds_rebin1.03" + HEN_FILE_EXTENSION,
        )
        cname = os.path.join(
            self.datadir,
            "monol_test_E3-50_cpds_rebin1.03" + HEN_FILE_EXTENSION,
        )
        hen.plot.main(
            [pname, cname, "--noplot", "--xlog", "--ylog", "-o", "dummy.qdp"]
        )
        hen.plot.main(
            [
                pname,
                "--noplot",
                "--axes",
                "power",
                "power_err",
                "--xlin",
                "--ylin",
                "-o",
                "dummy.qdp",
            ]
        )

    def test_plot_save_figure(self):
        """Test plotting and saving figure."""
        pname = os.path.join(
            self.datadir,
            "monol_testA_E3-50_pds_rebin1.03" + HEN_FILE_EXTENSION,
        )
        hen.plot.main(
            [
                pname,
                "--noplot",
                "--figname",
                os.path.join(
                    self.datadir, "monol_testA_E3-50_pds_rebin1.03.png"
                ),
                "-o",
                "dummy.qdp",
            ]
        )

    @classmethod
    def teardown_class(self):
        """Test a full run of the scripts (command lines)."""

        def find_file_pattern_in_dir(pattern, directory):
            return glob.glob(os.path.join(directory, pattern))

        patterns = [
            "*monol_test*" + HEN_FILE_EXTENSION,
            "*lcurve*" + HEN_FILE_EXTENSION,
            "*lcurve*.txt",
            "*.log",
            "*monol_test*.dat",
            "*monol_test*.png",
            "*monol_test*.txt",
            "*monol_test_fake*.evt",
            "*bubu*",
            "*.p",
            "*.qdp",
            "*.inf",
            "*_cpds" + HEN_FILE_EXTENSION,
            "*_ev" + HEN_FILE_EXTENSION,
        ]

        file_list = []
        for pattern in patterns:
            file_list.extend(find_file_pattern_in_dir(pattern, self.datadir))

        for f in file_list:
            if os.path.exists(f):
                print("Removing " + f)
                os.remove(f)

        patterns = ["*_pds*/", "*_cpds*/", "*_sum/"]

        dir_list = []
        for pattern in patterns:
            dir_list.extend(find_file_pattern_in_dir(pattern, self.datadir))
        for f in dir_list:
            if os.path.exists(f):
                shutil.rmtree(f)



"""Save different input files in PRESTO-readable format."""

from astropy import log
from astropy.coordinates import SkyCoord
import numpy as np
from .io import high_precision_keyword_read, get_file_type, HEN_FILE_EXTENSION
from .base import deorbit_events, interpret_bintime

MAXBIN = 100000000


def get_header_info(obj):
    """Get header info from a Stingray object."""
    from astropy.io.fits import Header

    header = Header.fromstring(obj.header)
    info = type("", (), {})()
    info.mjdref = high_precision_keyword_read(header, "MJDREF")
    info.telescope = header["TELESCOP"]
    info.instrument = header["INSTRUME"]
    info.source = header["OBJECT"]
    try:
        user = header["USER"]
    except KeyError:
        user = "Unknown"
    info.observer = user
    info.user = user
    info.tstart = header["TSTART"]
    info.tstop = header["TSTOP"]
    try:
        ra = header["RA_OBJ"]
        dec = header["DEC_OBJ"]
    except KeyError:
        ra = header["RA_PNT"]
        dec = header["DEC_PNT"]

    a = SkyCoord(ra, dec, unit="degree")
    info.raj = (
        (a.ra.to_string("hourangle"))
        .replace("s", "")
        .replace("h", ":")
        .replace("m", ":")
    )
    info.decj = (
        (a.ra.to_string()).replace("s", "").replace("d", ":").replace("m", ":")
    )
    if hasattr(obj, "e_interval"):
        e0, e1 = obj.e_interval
    elif hasattr(obj, "energy") and obj.energy is not None:
        e0, e1 = np.min(obj.energy), np.max(obj.energy)
    else:
        e0, e1 = 0, 0
    info.centralE = (e0 + e1) / 2
    info.bandpass = e1 - e0

    return info


def _save_to_binary(lc, filename):
    """Save a light curve to binary format."""
    nc = len(lc.counts)
    lc.counts[: nc // 2 * 2].astype("float32").tofile(filename)
    return


def save_lc_to_binary(lc, filename):
    """Save a light curve to binary format.

    Parameters
    ----------
    lc : `:class:stingray.Lightcurve`
        Input light curve
    filename : str
        Output file name
    Returns
    -------
    lcinfo : object
        light curve info
    """
    tstart = lc.tstart
    tstop = lc.tstart + lc.tseg
    nbin = lc.n
    bin_time = lc.dt
    _save_to_binary(lc, filename)

    lcinfo = type("", (), {})()
    lcinfo.bin_intervals_start = [0]
    lcinfo.bin_intervals_stop = [nbin]
    lcinfo.lclen = nbin
    lcinfo.tstart = tstart
    lcinfo.dt = bin_time
    lcinfo.tseg = tstop - tstart
    lcinfo.nphot = np.sum(lc.counts)
    return lcinfo


def save_events_to_binary(
    events, filename, bin_time, tstart=None, emin=None, emax=None
):
    """Save an event list to binary format.

    Parameters
    ----------
    events : `:class:stingray.Eventlist`
        Input event list
    filename : str
        Output file name
    bin_time : float
        Bin time of the output light curve

    Other parameters
    ----------------
    tstart : float
        Starting time
    emin : float
        Minimum energy of the photons
    emax : float
        Maximum energy of the photons

    Returns
    -------
    lcinfo : object
        light curve info
    """
    import struct

    if tstart is None:
        tstart = events.gti[0, 0]

    if emin is not None and emax is not None:
        if not hasattr(events, "energy") or events.energy is None:
            raise ValueError(
                "Energy filtering requested for uncalibrated event " "list"
            )

        good = (events.energy >= emin) & (events.energy < emax)
        events.time = events.time[good]

    tstop = events.gti[-1, 1]
    nbin = (tstop - tstart) / bin_time

    lclen = 0
    file = open(filename, "wb")
    nphot = 0
    for i in np.arange(0, nbin, MAXBIN):
        t0 = i * bin_time + tstart

        lastbin = int(np.min([MAXBIN, (nbin - i) // 4 * 4]))
        t1 = t0 + lastbin * bin_time

        good = (events.time >= t0) & (events.time < t1)

        goodev = events.time[good]
        hist, times = np.histogram(
            goodev, bins=np.linspace(t0, t1, lastbin + 1)
        )

        lclen += lastbin
        s = struct.pack("f" * len(hist), *hist)
        file.write(s)
        nphot += len(goodev)
    file.close()

    lcinfo = type("", (), {})()
    lcinfo.bin_intervals_start = np.floor(
        (events.gti[:, 0] - tstart) / bin_time
    )
    lcinfo.bin_intervals_stop = np.floor(
        (events.gti[:, 1] - tstart) / bin_time
    )
    lcinfo.lclen = lclen
    lcinfo.tstart = tstart
    lcinfo.dt = bin_time
    lcinfo.tseg = tstop - tstart
    lcinfo.nphot = nphot
    return lcinfo


def save_inf(lcinfo, info, filename):
    """Save information file."""

    lclen = lcinfo.lclen
    bin_intervals_start, bin_intervals_stop = (
        lcinfo.bin_intervals_start,
        lcinfo.bin_intervals_stop,
    )

    epoch = info.mjdref + lcinfo.tstart / 86400

    with open(filename, "w") as f:
        print(
            " Data file name without suffix         "
            " =  {}".format(filename.replace(".inf", "")),
            file=f,
        )
        print(
            " Telescope used                        "
            " =  {}".format(info.telescope),
            file=f,
        )
        print(
            " Instrument used                       "
            " =  {}".format(info.instrument),
            file=f,
        )
        print(
            " Object being observed                 "
            " =  {}".format(info.source),
            file=f,
        )
        print(
            " J2000 Right Ascension (hh:mm:ss.ssss) "
            " =  {}".format(info.raj),
            file=f,
        )
        print(
            " J2000 Declination     (dd:mm:ss.ssss) "
            " =  {}".format(info.decj),
            file=f,
        )
        print(
            " Data observed by                      "
            " =  {}".format(info.observer),
            file=f,
        )
        print(
            " Epoch of observation (MJD)            "
            " =  {:05.15f}".format(epoch),
            file=f,
        )
        print(" Barycentered?           (1=yes, 0=no) " " =  1", file=f)
        print(
            " Number of bins in the time series     "
            " =  {lclen}".format(lclen=lclen),
            file=f,
        )
        print(
            " Width of each time series bin (sec)   "
            " =  {bintime}".format(bintime=lcinfo.dt),
            file=f,
        )
        print(" Any breaks in the data? (1 yes, 0 no) " " =  1", file=f)
        for i, st in enumerate(bin_intervals_start):
            print(
                " On/Off bin pair # {ngti:>2}                  "
                " =  {binstart:<11}, "
                "{binstop:<11}".format(
                    ngti=i + 1, binstart=st, binstop=bin_intervals_stop[i]
                ),
                file=f,
            )
        print(" Type of observation (EM band)         " " =  X-ray", file=f)
        print(" Field-of-view diameter (arcsec)       " " =  400", file=f)
        print(
            " Central energy (kev)                  "
            " =  {}".format(info.centralE),
            file=f,
        )
        print(
            " Energy bandpass (kev)                 "
            " =  {}".format(info.bandpass),
            file=f,
        )
        print(
            " Data analyzed by                      "
            " =  {}".format(info.user),
            file=f,
        )
        print(" Any additional notes:", file=f)
        print(
            "       T = {length}, Nphot={nphot}".format(
                length=lcinfo.tseg, nphot=lcinfo.nphot
            ),
            file=f,
        )

    return


def main_presto(args=None):
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = "Save light curves in a format readable to PRESTO"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of input light curves", nargs="+")

    parser.add_argument(
        "-l",
        "--max-length",
        help="Maximum length of light " "curves (split otherwise)",
        type=np.longdouble,
        default=1e32,
    )

    args = check_negative_numbers_in_args(args)
    _add_default_args(
        parser,
        ["bintime", "energies", "deorbit", "nproc", "loglevel", "debug"],
    )

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)

    bintime = np.longdouble(interpret_bintime(args.bintime))

    if args.energy_interval is None:
        args.energy_interval = [None, None]
    with log.log_to_file("HENbinary.log"):
        for f in args.files:
            print(f)
            outfile = f.replace(HEN_FILE_EXTENSION, ".dat")
            ftype, contents = get_file_type(f)
            if ftype == "lc":
                lcinfo = save_lc_to_binary(contents, outfile)
            elif ftype == "events":
                if args.deorbit_par is not None:
                    contents = deorbit_events(contents, args.deorbit_par)
                lcinfo = save_events_to_binary(
                    contents,
                    outfile,
                    bin_time=bintime,
                    emin=args.energy_interval[0],
                    emax=args.energy_interval[1],
                )
            else:
                raise ValueError("File type not recognized")

            info = get_header_info(contents)
            save_inf(lcinfo, info, f.replace(HEN_FILE_EXTENSION, ".inf"))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Read and save event lists from FITS files."""

import warnings
import copy
import os
import numpy as np
from astropy import log
from stingray.io import load_events_and_gtis
from stingray.events import EventList
from stingray.gti import cross_two_gtis
from .io import load_events
from .base import common_name
from .base import hen_root
from .io import save_events
from .io import HEN_FILE_EXTENSION


def treat_event_file(
    filename,
    noclobber=False,
    gti_split=False,
    min_length=4,
    gtistring=None,
    length_split=None,
    randomize_by=None,
    discard_calibration=False,
    additional_columns=None,
):
    """Read data from an event file, with no external GTI information.

    Parameters
    ----------
    filename : str

    Other Parameters
    ----------------
    noclobber: bool
        if a file is present, do not overwrite it
    gtistring: str
        comma-separated set of GTI strings to consider
    gti_split: bool
        split the file in multiple chunks, containing one GTI each
    length_split: float, default None
        split the file in multiple chunks, with approximately this length
    min_length: float
        minimum length of GTIs accepted (only if gti_split is True or
        length_split is not None)
    discard_calibration: bool
        discard the automatic calibration done by Stingray (if any)
    """
    # gtistring = assign_value_if_none(gtistring, "GTI,GTI0,STDGTI")
    log.info("Opening %s" % filename)

    events = EventList.read(
        filename,
        format_="hea",
        gtistring=gtistring,
        additional_columns=additional_columns,
    )

    if discard_calibration:
        events.energy = None
        events.cal_pi = None

    mission = events.mission
    instr = events.instr.lower()
    gtis = events.gti
    lengths = np.array([g1 - g0 for (g0, g1) in gtis])
    gtis = gtis[lengths >= min_length]
    events.gti = gtis
    detector_id = events.detector_id

    if randomize_by is not None:
        events.time += np.random.uniform(
            -randomize_by / 2, randomize_by / 2, events.time.size
        )

    if detector_id is not None:
        detectors = np.array(list(set(detector_id)))
    else:
        detectors = [None]
    outfile_root = (
        hen_root(filename) + "_" + mission.lower() + "_" + instr.lower()
    )

    if randomize_by is not None:
        outfile_root += f"_randomize_by_{randomize_by:g}s"

    output_files = []
    for d in detectors:
        if d is not None:
            good_det = d == detector_id
            outroot_local = "{0}_det{1:02d}".format(outfile_root, d)

        else:
            good_det = np.ones_like(events.time, dtype=bool)
            outroot_local = outfile_root

        outfile = outroot_local + "_ev" + HEN_FILE_EXTENSION
        if (
            noclobber
            and os.path.exists(outfile)
            and (not (gti_split or length_split))
        ):
            warnings.warn(
                "{0} exists and using noclobber. Skipping".format(outfile)
            )
            return

        if gti_split or (length_split is not None):
            if length_split:
                gti0 = np.arange(gtis[0, 0], gtis[-1, 1], length_split)
                gti1 = gti0 + length_split
                gti_chunks = np.array(
                    [[g0, g1] for (g0, g1) in zip(gti0, gti1)]
                )
                label = "chunk"
            else:
                gti_chunks = gtis
                label = "gti"

            for ig, g in enumerate(gti_chunks):
                outfile_local = (
                    "{0}_{1}{2:03d}_ev".format(outroot_local, label, ig)
                    + HEN_FILE_EXTENSION
                )

                good_gtis = cross_two_gtis([g], gtis)
                if noclobber and os.path.exists(outfile_local):
                    warnings.warn(
                        "{0} exists, ".format(outfile_local)
                        + "and noclobber option used. Skipping"
                    )
                    return
                good = np.logical_and(events.time >= g[0], events.time < g[1])
                all_good = good_det & good
                if len(events.time[all_good]) < 1:
                    continue
                events_filt = events.apply_mask(all_good)
                events_filt.gti = good_gtis

                save_events(events_filt, outfile_local)
                output_files.append(outfile_local)
        else:
            events_filt = events.apply_mask(good_det)

            save_events(events_filt, outfile)
            output_files.append(outfile)
    return output_files


def _wrap_fun(arglist):
    f, kwargs = arglist
    try:
        return treat_event_file(f, **kwargs)
    except IndexError:
        log.error(f"Empty or corrupt event file: {f}")
    except Exception as e:
        log.error(f"Unknown error: {f}")
        log.error(f"{str(e)}")


def multiple_event_concatenate(event_lists):
    """
    Join multiple :class:`EventList` objects into one.

    If both are empty, an empty :class:`EventList` is returned.

    GTIs are crossed if the event lists are over a common time interval,
    and appended otherwise.

    ``pi`` and ``pha`` remain ``None`` if they are ``None`` in both.
    Otherwise, 0 is used as a default value for the :class:`EventList` where
    they were None.

    Parameters
    ----------
    event_lists : list of :class:`EventList` object
        :class:`EventList` objects that we are joining

    Returns
    -------
    `ev_new` : :class:`EventList` object
        The resulting :class:`EventList` object.
    """

    ev_new = EventList()

    gtis = np.concatenate([ev.gti for ev in event_lists])
    order = np.argsort(gtis[:, 0])
    gtis = gtis[order]

    ev_new.time = np.concatenate([ev.time for ev in event_lists])
    order = np.argsort(ev_new.time)
    ev_new.time = ev_new.time[order]

    if hasattr(event_lists[0], "pi") and event_lists[0].pi is not None:
        ev_new.pi = np.concatenate([ev.pi for ev in event_lists])[order]
    if hasattr(event_lists[0], "energy") and event_lists[0].energy is not None:
        ev_new.energy = np.concatenate([ev.energy for ev in event_lists])[
            order
        ]

    ev_new.mjdref = event_lists[0].mjdref
    ev_new.gti = gtis

    return ev_new


def join_eventlists(event_file1, event_file2, new_event_file=None):
    """Join two event files.

    Parameters
    ----------
    event_file1 : str
        First event file
    event_file2 : str
        Second event file

    Other parameters
    ----------------
    new_event_file : str, default None
        Output event file. If not specified uses `hendrics.utils.common_name`
        to figure out a good name to use mixing up the two input names.

    Returns
    -------
    new_event_file : str
        Output event file
    """
    if new_event_file is None:
        new_event_file = (
            common_name(event_file1, event_file2) + "_ev" + HEN_FILE_EXTENSION
        )

    events1 = load_events(event_file1)
    events2 = load_events(event_file2)
    if events2.time.size == 0 or events2.gti.size == 0:
        warnings.warn(f"{event_file2} has no good events")
        return None

    if events2.mjdref != events1.mjdref:
        warnings.warn("Different missions detected; changing MJDREF")
        time_diff = (events1.mjdref - events2.mjdref) * 86400
        events2.time -= time_diff
        events2.mjdref = events1.mjdref
        events2.gti -= time_diff

    events = events1.join(events2)
    if hasattr(events2, "header"):
        events.header = events1.header
    if events1.instr.lower() != events2.instr.lower():
        events.instr = ",".join([e.instr.lower() for e in [events1, events2]])
    for attr in ["mission", "instr"]:
        if getattr(events1, attr) != getattr(events2, attr):
            setattr(
                events,
                attr,
                getattr(events1, attr) + "," + getattr(events2, attr),
            )
        else:
            setattr(events, attr, getattr(events1, attr))

    save_events(events, new_event_file)

    return new_event_file


def join_many_eventlists(eventfiles, new_event_file=None):
    """Join two event files.

    Parameters
    ----------
    event_files : list of str
        List of event files

    Other parameters
    ----------------
    new_event_file : str, default None
        Output event file. If not specified ``joint_ev`` + HEN_FILE_EXTENSION

    Returns
    -------
    new_event_file : str
        Output event file
    """
    if new_event_file is None:
        new_event_file = "joint_ev" + HEN_FILE_EXTENSION

    N = len(eventfiles)
    first_events = load_events(eventfiles[0])
    all_events = [first_events]
    for i, event_file in enumerate(eventfiles[1:]):
        log.info(f"Reading {event_file} ({i + 1}/{N})")
        events = load_events(event_file)
        if not np.isclose(events.mjdref, first_events.mjdref):
            warnings.warn(f"{event_file} has a different MJDREF")
            continue
        if hasattr(events, "instr") and not events.instr == first_events.instr:
            warnings.warn(f"{event_file} is from a different instrument")
            continue
        if (
            events.time.size == 0
            or events.gti.size == 0
            or not np.all(
                [
                    events.time[0] < events.gti.max(),
                    events.time[-1] > events.gti.min(),
                ]
            )
        ):
            warnings.warn(f"{event_file} has no good events")
            continue

        all_events.append(events)

    events = multiple_event_concatenate(all_events)
    save_events(events, new_event_file)
    return new_event_file


def _split_events(ev, max_length, overlap=0):
    event_times = ev.time
    GTI = ev.gti
    t0 = GTI[0, 0]
    while t0 < GTI.max():
        t1 = min(t0 + max_length, GTI.max())
        if t1 - t0 < max_length / 2:
            break
        idx_start = np.searchsorted(event_times, t0)
        idx_stop = np.searchsorted(event_times, t1)
        gti_local = cross_two_gtis(GTI, [[t0, t1]])

        local_times = event_times[idx_start:idx_stop]
        new_ev = EventList(time=local_times, gti=gti_local)
        for attr in ["pi", "energy", "cal_pi"]:
            if hasattr(ev, attr) and getattr(ev, attr) is not None:
                setattr(new_ev, attr, getattr(ev, attr)[idx_start:idx_stop])
        for attr in ["mission", "instr", "mjdref", "header"]:
            if hasattr(ev, attr) and getattr(ev, attr) is not None:
                setattr(new_ev, attr, getattr(ev, attr))
        t0 = t0 + max_length * (1.0 - overlap)
        yield new_ev


def split_eventlist(fname, max_length, overlap=None):
    root = hen_root(fname)
    ev = load_events(fname)

    if overlap is None:
        overlap = 0
    if overlap >= 1:
        raise ValueError("Overlap cannot be >=1. Exiting.")

    event_times = ev.time
    GTI = ev.gti
    t0 = GTI[0, 0]
    count = 0
    from .base import nchars_in_int_value

    nchars = nchars_in_int_value((GTI.max() - t0) / max_length)

    all_files = []
    for new_ev in _split_events(ev, max_length, overlap):
        newfname = root + f"_{count:0{nchars}d}" + HEN_FILE_EXTENSION

        save_events(new_ev, newfname)
        all_files.append(newfname)
        count += 1
    return all_files


def main_join(args=None):
    """Main function called by the `HENjoinevents` command line script."""
    import argparse

    description = (
        "Read a cleaned event files and saves the relevant "
        "information in a standard format"
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="Files to join", type=str, nargs="+")
    parser.add_argument(
        "-o", "--output", type=str, help="Name of output file", default=None
    )
    args = parser.parse_args(args)

    if len(args.files) == 2:
        return join_eventlists(
            args.files[0], args.files[1], new_event_file=args.output
        )
    else:
        return join_many_eventlists(args.files, new_event_file=args.output)


def main_splitevents(args=None):
    """Main function called by the `HENsplitevents` command line script."""
    import argparse

    description = (
        "Reads a cleaned event files and splits the file into "
        "overlapping multiple chunks of fixed length"
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("fname", help="File 1", type=str)
    parser.add_argument(
        "-l",
        "--length-split",
        help="Split event list by GTI",
        default=None,
        type=float,
    )
    parser.add_argument(
        "--overlap",
        type=float,
        help="Overlap factor. 0 for no overlap, 0.5 for "
        "half-interval overlap, and so on.",
        default=None,
    )

    args = parser.parse_args(args)

    return split_eventlist(
        args.fname, max_length=args.length_split, overlap=args.overlap
    )


def main(args=None):
    """Main function called by the `HENreadevents` command line script."""
    import argparse
    from multiprocessing import Pool
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Read a cleaned event files and saves the relevant "
        "information in a standard format"
    )
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "--noclobber",
        help=("Do not overwrite existing event files"),
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-g",
        "--gti-split",
        help="Split event list by GTI",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--discard-calibration",
        help="Discard automatic calibration (if any)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-l",
        "--length-split",
        help="Split event list by length",
        default=None,
        type=float,
    )
    parser.add_argument(
        "--min-length",
        type=int,
        help="Minimum length of GTIs to consider",
        default=0,
    )
    parser.add_argument(
        "--gti-string", type=str, help="GTI string", default=None
    )
    parser.add_argument(
        "--randomize-by",
        type=float,
        help="Randomize event arrival times by this amount "
        "(e.g. it might be the 0.073-s frame time in "
        "XMM)",
        default=None,
    )
    parser.add_argument(
        "--additional",
        type=str,
        nargs="+",
        help="Additional columns to be read from the FITS file",
        default=None,
    )
    _add_default_args(parser, ["output", "loglevel", "debug", "nproc"])

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)

    with log.log_to_file("HENreadevents.log"):
        argdict = {
            "noclobber": args.noclobber,
            "gti_split": args.gti_split,
            "min_length": args.min_length,
            "gtistring": args.gti_string,
            "length_split": args.length_split,
            "randomize_by": args.randomize_by,
            "discard_calibration": args.discard_calibration,
            "additional_columns": args.additional,
        }

        arglist = [[f, argdict] for f in files]

        if os.name == "nt" or args.nproc == 1:
            [_wrap_fun(a) for a in arglist]
        else:
            pool = Pool(processes=args.nproc)
            for i in pool.imap_unordered(_wrap_fun, arglist):
                pass
            pool.close()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Calibrate event lists by looking in rmf files."""

import os
import warnings
import numpy as np
from astropy import log
from .io import load_events, save_events, get_file_extension
from .io import HEN_FILE_EXTENSION


def default_nustar_rmf():
    """Look for the default rmf file in the CALDB.

    The CALDB environment variable has to point to the correct location of
    the NuSTAR CALDB

    .. note:: The calibration might change in the future. The hardcoded file
              name will be eventually replaced with a smarter choice based
              on observing time
    """
    warnings.warn("Rmf not specified. Using default NuSTAR rmf.")
    rmf = "data/nustar/fpm/cpf/rmf/nuAdet3_20100101v002.rmf"
    path = rmf.split("/")
    newpath = os.path.join(os.environ["CALDB"], *path)
    return newpath


def read_rmf(rmf_file=None):
    """Load RMF info.

    .. note:: Preliminary: only EBOUNDS are read.

    Parameters
    ----------
    rmf_file : str
        The rmf file used to read the calibration. If None or not specified,
        the one given by default_nustar_rmf() is used.

    Returns
    -------
    pis : array-like
        the PI channels
    e_mins : array-like
        the lower energy bound of each PI channel
    e_maxs : array-like
        the upper energy bound of each PI channel
    """
    from astropy.io import fits as pf

    if rmf_file is None or rmf_file == "":
        rmf_file = default_nustar_rmf()

    lchdulist = pf.open(rmf_file, checksum=True)
    lchdulist.verify("warn")
    lctable = lchdulist["EBOUNDS"].data
    pis = np.array(lctable.field("CHANNEL"))
    e_mins = np.array(lctable.field("E_MIN"))
    e_maxs = np.array(lctable.field("E_MAX"))
    lchdulist.close()
    return pis, e_mins, e_maxs


def read_calibration(pis, rmf_file=None):
    """Read the energy channels corresponding to the given PI channels.

    Parameters
    ----------
    pis : array-like
        The channels to lookup in the rmf

    Other Parameters
    ----------------
    rmf_file : str
        The rmf file used to read the calibration. If None or not specified,
        the one given by default_nustar_rmf() is used.
    """
    calp, calEmin, calEmax = read_rmf(rmf_file)
    es = np.zeros(len(pis), dtype=float)
    for ic, c in enumerate(calp):
        good = pis == c
        if not np.any(good):
            continue
        es[good] = (calEmin[ic] + calEmax[ic]) / 2

    return es


def rough_calibration(pis, mission):
    """

    Parameters
    ----------
    pis: float or array of floats
        PI channels in data
    mission: str
        Mission name

    Returns
    -------
    energies : float or array of floats
        Energy values

    Examples
    --------
    >>> rough_calibration(0, 'nustar')
    1.6
    >>> # It's case-insensitive
    >>> rough_calibration(1200, 'XMm')
    1.2
    >>> rough_calibration(10, 'asDf')
    Traceback (most recent call last):
        ...
    ValueError: Mission asdf not recognized
    >>> rough_calibration(100, 'nicer')
    1.0
    """
    if mission.lower() == "nustar":
        return pis * 0.04 + 1.6
    elif mission.lower() == "xmm":
        return pis * 0.001
    elif mission.lower() == "nicer":
        return pis * 0.01
    raise ValueError(f"Mission {mission.lower()} not recognized")


def calibrate(fname, outname, rmf_file=None, rough=False):
    """Do calibration of an event list.

    Parameters
    ----------
    fname : str
        The HENDRICS file containing the events
    outname : str
        The output file

    Other Parameters
    ----------------
    rmf_file : str
        The rmf file used to read the calibration. If None or not specified,
        the one given by default_nustar_rmf() is used.
    """
    # Read event file
    log.info("Loading file %s..." % fname)
    evdata = load_events(fname)
    log.info("Done.")
    pis = evdata.pi

    if rough:
        cal_pis = evdata.pi
        if hasattr(evdata, "cal_pi") and evdata.cal_pi is not None:
            cal_pis = evdata.cal_pi
        es = rough_calibration(cal_pis, evdata.mission)
    else:
        es = read_calibration(pis, rmf_file)

    evdata.energy = es
    log.info("Saving calibrated data to %s" % outname)
    save_events(evdata, outname)


def _calib_wrap(args):
    f, outname, rmf, rough = args
    return calibrate(f, outname, rmf, rough)


def main(args=None):
    """Main function called by the `HENcalibrate` command line script."""
    import argparse
    from multiprocessing import Pool
    from .base import _add_default_args

    description = (
        "Calibrate clean event files by associating the correct "
        "energy to each PI channel. Uses either a specified rmf "
        "file or (for NuSTAR only) an rmf file from the CALDB"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "-r",
        "--rmf",
        help="rmf file used for calibration",
        default=None,
        type=str,
    )
    parser.add_argument(
        "--rough",
        help="Rough calibration, without rmf file "
        "(only for NuSTAR and XMM)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-o",
        "--overwrite",
        help="Overwrite; default: no",
        default=False,
        action="store_true",
    )
    _add_default_args(parser, ["nproc", "loglevel", "debug"])

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENcalibrate.log"):
        funcargs = []
        for i_f, f in enumerate(files):
            outname = f
            if args.overwrite is False:
                label = "_calib"
                if args.rough:
                    label = "_rough_calib"
                outname = f.replace(
                    get_file_extension(f), label + HEN_FILE_EXTENSION
                )
            funcargs.append([f, outname, args.rmf, args.rough])

        if os.name == "nt" or args.nproc == 1:
            [_calib_wrap(fa) for fa in funcargs]
        else:
            pool = Pool(processes=args.nproc)
            for i in pool.imap_unordered(_calib_wrap, funcargs):
                pass
            pool.close()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to simulate data and produce a fake event file."""

import os
import warnings
import copy
import numpy as np
import numpy.random as ra
from astropy import log
from astropy.io.fits import Header
from astropy.logger import AstropyUserWarning
from stingray.events import EventList
from stingray.lightcurve import Lightcurve
from stingray.utils import assign_value_if_none
from stingray.filters import filter_for_deadtime
from stingray.io import read_mission_info, get_key_from_mission_info
from .io import get_file_format, load_lcurve
from .io import load_events, save_events, HEN_FILE_EXTENSION
from .base import _empty, r_in
from .fold import filter_energy
from .lcurve import lcurve_from_fits
from .base import njit, deorbit_events


def _clean_up_header(header):
    if header is None:
        return None
    for key in header.keys():
        for k in ["TTYP", "TFORM"]:
            if key.startswith(k):
                header.pop(key)
    for k in ["EXTNAME", "PCOUNT", "GCOUNT", "NAXIS1", "NAXIS2"]:
        if k in header:
            header.pop(k)
    return header


def _fill_in_default_information(tbheader):
    tbheader["OBSERVER"] = "Edwige Bubble"
    tbheader["COMMENT"] = (
        "FITS (Flexible Image Transport System) format is"
        " defined in 'Astronomy and Astrophysics', volume"
        " 376, page 359; bibcode: 2001A&A...376..359H"
    )
    tbheader["OBS_ID"] = ("00000000001", "Observation ID")
    tbheader["TARG_ID"] = (0, "Target ID")
    tbheader["OBJECT"] = ("Fake X-1", "Name of observed object")
    tbheader["RA_OBJ"] = (0.0, "[deg] R.A. Object")
    tbheader["DEC_OBJ"] = (0.0, "[deg] Dec Object")
    tbheader["RA_NOM"] = (
        0.0,
        "Right Ascension used for barycenter corrections",
    )
    tbheader["DEC_NOM"] = (0.0, "Declination used for barycenter corrections")
    tbheader["RA_PNT"] = (0.0, "[deg] RA pointing")
    tbheader["DEC_PNT"] = (0.0, "[deg] Dec pointing")
    tbheader["PA_PNT"] = (0.0, "[deg] Position angle (roll)")
    tbheader["EQUINOX"] = (2.000e03, "Equinox of celestial coord system")
    tbheader["RADECSYS"] = ("FK5", "Coordinate Reference System")
    tbheader["TASSIGN"] = ("SATELLITE", "Time assigned by onboard clock")
    tbheader["TIMESYS"] = ("TDB", "All times in this file are TDB")
    tbheader["TIMEREF"] = (
        "SOLARSYSTEM",
        "Times are pathlength-corrected to barycenter",
    )
    tbheader["CLOCKAPP"] = (False, "TRUE if timestamps corrected by gnd sware")
    tbheader["COMMENT"] = (
        "MJDREFI+MJDREFF = epoch of Jan 1, 2010, in TT " "time system."
    )
    tbheader["TIMEUNIT"] = ("s", "unit for time keywords")
    return tbheader


def generate_fake_fits_observation(
    event_list=None,
    filename=None,
    instr=None,
    gti=None,
    tstart=None,
    tstop=None,
    mission=None,
    mjdref=55197.00076601852,
    livetime=None,
    additional_columns={},
):
    """Generate fake X-ray data.

    Takes an event list (as a list of floats)
    All inputs are None by default, and can be set during the call.

    Parameters
    ----------
    event_list : list-like
        :class:`stingray.events.Eventlist` object. If left None, 1000
        random events will be generated, for a total length of 1025 s or the
        difference between tstop and tstart.
    filename : str
        Output file name

    Returns
    -------
    hdulist : FITS hdu list
        FITS hdu list of the output file

    Other Parameters
    ----------------
    mjdref : float
        Reference MJD. Default is 55197.00076601852 (NuSTAR)
    pi : list-like
        The PI channel of each event
    tstart : float
        Start of the observation (s from mjdref)
    tstop : float
        End of the observation (s from mjdref)
    instr : str
        Name of the instrument. Default is 'FPMA'
    livetime : float
        Total livetime. Default is tstop - tstart
    """
    from astropy.io import fits
    import numpy.random as ra

    inheader = None
    if event_list is None:
        tstart = assign_value_if_none(tstart, 8e7)
        tstop = assign_value_if_none(tstop, tstart + 1025)
        ev_list = np.sort(ra.uniform(tstart, tstop, 1000))
        gti = assign_value_if_none(gti, np.array([[tstart, tstop]]))
    else:
        if hasattr(event_list, "header") and event_list.header is not None:
            inheader = Header.fromstring(event_list.header)
            inheader = _clean_up_header(inheader)

        ev_list = event_list.time
        gti = assign_value_if_none(
            event_list.gti, np.asarray([[ev_list[0], ev_list[-1]]])
        )
        mission = assign_value_if_none(mission, event_list.mission)
        instr = assign_value_if_none(instr, event_list.instr)
        tstart = assign_value_if_none(tstart, gti[0, 0])
        tstop = assign_value_if_none(tstop, gti[-1, 1])
        if hasattr(event_list, "mjdref") and event_list.mjdref is not None:
            mjdref = event_list.mjdref

    mission = assign_value_if_none(mission, "NuSTAR")
    instr = assign_value_if_none(instr, "FPMA")
    if hasattr(event_list, "pi") and event_list.pi is not None:
        pi = event_list.pi
    else:
        pi = ra.randint(0, 1024, np.size(ev_list))

    if hasattr(event_list, "cal_pi") and event_list.cal_pi is not None:
        cal_pi = event_list.cal_pi
    else:
        cal_pi = pi / 3

    filename = assign_value_if_none(filename, "events.evt")
    livetime = assign_value_if_none(livetime, tstop - tstart)

    if livetime > tstop - tstart:
        raise ValueError(
            "Livetime must be equal or smaller than " "tstop - tstart"
        )

    mission_info = read_mission_info(mission)
    allowed_instr = mission_info["instruments"]

    # Just prefer EPN for XMM
    if "xmm" in mission.lower():
        allowed_instr = ["EPN", "EMOS1", "EMOS2", "RGS1", "RGS2"]

    allowed_instr = [ins.lower() for ins in allowed_instr]
    if instr.lower() not in allowed_instr:
        instr = allowed_instr[0]

    ccol = get_key_from_mission_info(mission_info, "ccol", None, inst=instr)
    if ccol.lower() == "none":
        ccol = None
    ecol = get_key_from_mission_info(mission_info, "ecol", "PI", inst=instr)
    ext = get_key_from_mission_info(
        mission_info, "events", "EVENTS", inst=instr
    )

    # Create primary header
    prihdr = fits.Header()
    prihdr["OBSERVER"] = "Edwige Bubble"
    if inheader is not None and "OBSERVER" in inheader:
        prihdr["OBSERVER"] = inheader["OBSERVER"]
    prihdr["TELESCOP"] = (mission, "Telescope (mission) name")
    prihdr["INSTRUME"] = (instr, "Instrument name")
    prihdu = fits.PrimaryHDU(header=prihdr)
    prihdu.verify("exception")

    # Write events to table
    col1 = fits.Column(name="TIME", format="1D", array=ev_list)

    allcols = [col1]

    if ccol is not None:
        if (
            not hasattr(event_list, "detector_id")
            or event_list.detector_id is None
        ):

            ccdnr = np.zeros(np.size(ev_list)) + 1
            ccdnr[1] = 2  # Make it less trivial
            ccdnr[10] = 7
        else:
            ccdnr = event_list.detector_id

        allcols.append(fits.Column(name=ccol, format="1J", array=ccdnr))

    if mission.lower().strip() in ["xmm", "swift"]:
        allcols.append(fits.Column(name="PHA", format="1J", array=pi))
        allcols.append(fits.Column(name="PI", format="1J", array=cal_pi))
    else:
        allcols.append(fits.Column(name=ecol, format="1J", array=pi))

    for c in additional_columns.keys():
        col = fits.Column(
            name=c,
            array=additional_columns[c]["data"],
            format=additional_columns[c]["format"],
        )
        allcols.append(col)

    cols = fits.ColDefs(allcols)

    # ---- Fake lots of information ----

    tbheader = Header()
    tbheader = _fill_in_default_information(tbheader)
    # If None, it will not update
    tbheader.update(inheader)

    tbheader["TSTART"] = (
        tstart,
        "Elapsed seconds since MJDREF at start of file",
    )
    tbheader["TELESCOP"] = (mission, "Telescope (mission) name")
    tbheader["INSTRUME"] = (instr, "Instrument name")
    if mjdref != float(int(mjdref)):
        tbheader["MJDREFI"] = (
            int(mjdref),
            "TDB time reference; Modified Julian Day (int)",
        )
        tbheader["MJDREFF"] = (
            mjdref - int(mjdref),
            "TDB time reference; Modified Julian Day (frac)",
        )
        tbheader.pop("MJDREF", None)
    else:
        tbheader["MJDREF"] = mjdref
        tbheader.pop("MJDREFI", None)
        tbheader.pop("MJDREFF", None)

    tbheader["TSTOP"] = (tstop, "Elapsed seconds since MJDREF at end of file")
    tbheader["LIVETIME"] = (livetime, "On-source time")
    tbheader["TIMEZERO"] = (0.000000e00, "Time Zero")
    tbheader["HISTORY"] = "Generated with HENDRICS by {0}".format(
        os.getenv("USER")
    )

    tbhdu = fits.BinTableHDU.from_columns(cols, header=tbheader)
    tbhdu.name = ext

    tbhdu.add_checksum()
    tbhdu.verify("exception")
    # ---- END Fake lots of information ----

    # Fake GTIs

    start = gti[:, 0]
    stop = gti[:, 1]

    col1 = fits.Column(name="START", format="1D", array=start)
    col2 = fits.Column(name="STOP", format="1D", array=stop)
    allcols = [col1, col2]
    cols = fits.ColDefs(allcols)
    gtinames = ["GTI"]
    if mission.lower().strip() == "xmm":
        gtinames = []
        for i in set(ccdnr):
            gtinames.append(f"STDGTI{int(i):02d}")

    all_new_hdus = [prihdu, tbhdu]
    for name in gtinames:
        gtihdu = fits.BinTableHDU.from_columns(cols)
        gtihdu.name = name
        gtihdu.verify("exception")
        all_new_hdus.append(gtihdu)

    tbhdu.verify("exception")

    thdulist = fits.HDUList(all_new_hdus)
    assert thdulist[1].verify_datasum() == 1
    thdulist.writeto(
        filename, overwrite=True, checksum=True, output_verify="exception"
    )

    thdulist.close()
    return filename


def _read_event_list(filename):
    ev_list = load_events(filename)
    return ev_list


def _read_light_curve(filename):
    file_format = get_file_format(filename)
    if file_format == "fits":
        filename = lcurve_from_fits(filename)[0]
    lc = load_lcurve(filename)

    return lc


def acceptance_rejection(
    dt, counts_per_bin, t0=0.0, poissonize_n_events=False, deadtime=0.0
):
    """
    Examples
    --------
    >>> counts_per_bin = [10, 5, 5]
    >>> dt = 0.1
    >>> ev = acceptance_rejection(dt, counts_per_bin)
    >>> ev.size == 20
    True
    >>> ev.max() < 0.3
    True
    >>> ev.min() > 0
    True
    >>> np.all(np.diff(ev) >= 0)
    True
    """
    counts_per_bin = np.asarray(counts_per_bin)
    rates = counts_per_bin / dt
    dead_time_corrected_rates = r_in(deadtime, rates)
    counts_per_bin = dead_time_corrected_rates * dt

    n_events = np.rint(np.sum(counts_per_bin)).astype(int)
    if poissonize_n_events:
        n_events = np.random.poisson(n_events)

    n_bins = counts_per_bin.size
    event_times = np.zeros(n_events)
    n_missing = n_events
    M = np.max(counts_per_bin)

    while n_missing > 0:
        stats = np.random.uniform(0, M, n_missing)
        float_bin = np.random.uniform(0, n_bins, n_missing)

        int_bin = np.floor(float_bin).astype(int)
        good = stats < counts_per_bin[int_bin]
        n = np.count_nonzero(good)
        if n == 0:
            continue
        start_bin = -n_missing
        end_bin = -n_missing + n
        if end_bin == 0:
            end_bin = event_times.size

        event_times[start_bin:end_bin] = float_bin[good] * dt + t0
        n_missing -= n

    return filter_for_deadtime(np.sort(event_times), deadtime)


def make_counts_pulsed(nevents, t_start, t_stop, pulsed_fraction=0.0):
    """

    Examples
    --------
    >>> nevents = 10
    >>> dt, counts = make_counts_pulsed(nevents, 0, 100)
    >>> np.isclose(np.sum(counts), nevents)
    True
    >>> dt, counts = make_counts_pulsed(nevents, 0, 100, pulsed_fraction=1)
    >>> np.isclose(np.sum(counts), nevents)
    True
    """
    dt = 0.0546372810934756
    length = t_start - t_stop
    n_bins = int(np.ceil(length / dt))
    # make dt an exact divisor of the length
    dt = length / n_bins

    times = np.arange(t_start, t_stop, dt)
    sinusoid = pulsed_fraction / 2 * np.sin(np.pi * 2 * times)

    lc = 1 - pulsed_fraction / 2 + sinusoid

    counts = lc * nevents / np.sum(lc)
    return dt, counts


def scramble(
    event_list,
    smooth_kind="flat",
    dt=None,
    pulsed_fraction=0.0,
    deadtime=0.0,
    orbit_par=None,
    frequency=1,
):
    """Scramble event list, GTI by GTI.

    Parameters
    ----------
    event_list: :class:`stingray.events.Eventlist` object
        Input event list

    Other parameters
    ----------------
    smooth_kind: str in ['flat', 'smooth', 'pulsed']
        if 'flat', count the events GTI by GTI without caring about long-term
        variability; if 'smooth', try to calculate smooth light curve first
    dt: float
        If ``smooth_kind`` is 'smooth', bin the light curve with this bin time.
        Ignored for other values of ``smooth_kind``
    pulsed_fraction: float
        If ``smooth_kind`` is 'pulsed', use this pulse fraction, defined as the
        2 A / B, where A is the amplitude of the sinusoid and B the maximum
        flux. Ignored for other values of ``smooth_kind``
    deadtime: float
        Dead time in the data.
    orbit_par: str
        Parameter file for orbital modulation

    Returns
    -------
    new_event_list: :class:`stingray.events.Eventlist` object
        "Scrambled" event list

    Examples
    --------
    >>> times = np.array([0.5, 134, 246, 344, 867])
    >>> event_list = EventList(
    ...     times, gti=np.array([[0, 0.9], [111, 123.2], [125.123, 1000]]))
    >>> new_event_list = scramble(event_list, 'smooth')
    >>> new_event_list.time.size == times.size
    True
    >>> np.all(new_event_list.gti == event_list.gti)
    True
    >>> new_event_list = scramble(event_list, 'flat')
    >>> new_event_list.time.size == times.size
    True
    >>> np.all(new_event_list.gti == event_list.gti)
    True
    """
    new_event_list = copy.deepcopy(event_list)
    assert np.all(np.diff(new_event_list.time) > 0)

    idxs = np.searchsorted(new_event_list.time, new_event_list.gti)

    if smooth_kind == "pulsed":
        # Frequency is one, but can be anywhere in the frequency bin (for
        # sensitivity losses)
        length = new_event_list.gti.max() - new_event_list.gti.min()
        df = 0.5 / length
        frequency += np.random.uniform(-df, df)

    for (i_start, i_stop), gti_boundary in zip(idxs, new_event_list.gti):
        locally_flat = False
        nevents = i_stop - i_start
        t_start, t_stop = gti_boundary[0], gti_boundary[1]
        if nevents < 1:
            continue
        length = t_stop - t_start
        if nevents < 10 and smooth_kind == "pulsed":
            continue
        if length <= 1:
            # in very short GTIs, always assume a flat distribution.
            locally_flat = True

        if smooth_kind == "flat" or locally_flat:
            rate = nevents / length
            input_rate = r_in(deadtime, rate)
            new_events = np.sort(
                np.random.uniform(
                    t_start,
                    t_stop,
                    np.rint(nevents * input_rate / rate).astype(int),
                )
            )
            new_events = filter_for_deadtime(new_events, deadtime)
            new_event_list.time[i_start:i_stop] = new_events[
                : i_stop - i_start
            ]
            continue
        elif smooth_kind == "smooth":
            if dt is None:
                # Try to have at least 20 counts per bin on average
                dt = min(length / (nevents / 20), length)
            # make dt an exact divisor of the length
            n_bins = int(np.ceil(length / dt))
            dt = length / n_bins

            counts, _ = np.histogram(
                new_event_list.time[i_start:i_stop],
                range=[t_start, t_stop],
                bins=n_bins,
            )

        elif smooth_kind == "pulsed":
            # dt must be sufficiently small so that the frequency can be
            # detected with no loss of sensitivity. Moreover, not exactly a
            # multiple of the frequency, to increase randomness in the
            # detection sensitivity. I take a random Nyquist frequency between
            # 10 and 15 times the pulse frequency
            nyq = np.random.uniform(frequency * 10, frequency * 15)
            dt = 0.5 / nyq
            n_bins = int(np.ceil(length / dt))
            # make dt an exact divisor of the length
            dt = length / n_bins

            times = np.arange(t_start, t_stop, dt)
            sinusoid = (
                pulsed_fraction / 2 * np.sin(np.pi * 2 * times * frequency)
            )

            lc = 1 - pulsed_fraction / 2 + sinusoid

            counts = lc * nevents / np.sum(lc)
        else:
            raise ValueError("Unknown value for `smooth_kind`")

        newev = acceptance_rejection(
            dt,
            counts,
            t0=t_start,
            poissonize_n_events=False,
            deadtime=deadtime,
        )
        new_event_list.time[i_start:i_stop] = newev

    if orbit_par is not None:
        new_event_list = deorbit_events(new_event_list, orbit_par, invert=True)

    return new_event_list


def main_scramble(args=None):
    """Main function called by the `HENscramble` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Scramble the events inside an event list, maintaining the same "
        "energies and GTIs"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "fname",
        type=str,
        default=None,
        help="File containing input event list",
    )
    parser.add_argument(
        "--smooth-kind",
        choices=["smooth", "flat", "pulsed"],
        help="Special testing value",
        default="flat",
    )
    parser.add_argument(
        "--deadtime",
        type=float,
        default=0,
        help="Dead time magnitude. Can be specified as a "
        "single number, or two. In this last case, the "
        "second value is used as sigma of the dead time "
        "distribution",
    )
    parser.add_argument(
        "--dt",
        type=float,
        default=0,
        help="Time resolution of smoothed light curve",
    )
    parser.add_argument(
        "--pulsed-fraction",
        type=float,
        default=0,
        help="Pulsed fraction of simulated pulsations",
    )
    parser.add_argument(
        "-f",
        "--frequency",
        type=float,
        default=1,
        help="Pulsed fraction of simulated pulsations",
    )
    parser.add_argument(
        "--outfile", type=str, default=None, help="Output file name"
    )
    args = check_negative_numbers_in_args(args)
    _add_default_args(parser, ["deorbit", "energies", "loglevel", "debug"])

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)

    event_list = load_events(args.fname)
    emin = emax = None
    if args.energy_interval is not None:
        emin, emax = args.energy_interval
        event_list, elabel = filter_energy(event_list, emin, emax)
        if elabel != "Energy":
            raise ValueError(
                "You are filtering by energy but the data are not calibrated"
            )

    new_event_list = scramble(
        event_list,
        smooth_kind=args.smooth_kind,
        dt=args.dt,
        pulsed_fraction=args.pulsed_fraction,
        deadtime=args.deadtime,
        orbit_par=args.deorbit_par,
        frequency=args.frequency,
    )

    if args.outfile is not None:
        outfile = args.outfile
    else:
        label = "_scramble"

        if args.smooth_kind == "pulsed":
            label += f"_pulsed_df{args.pulsed_fraction:g}"
        elif args.smooth_kind == "smooth":
            label += f"_smooth_dt{args.dt:g}s"
        if args.deadtime > 0:
            label += f"_deadtime_{args.deadtime:g}"
        if args.energy_interval is not None:
            label += f"_{emin:g}-{emax:g}keV"

        outfile = args.fname.replace(
            HEN_FILE_EXTENSION, f"{label}" + HEN_FILE_EXTENSION
        )
    save_events(new_event_list, outfile)
    return outfile


def main(args=None):
    """Main function called by the `HENfake` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Create an event file in FITS format from an event list, or simulating"
        " it. If input event list is not specified, generates the events "
        "randomly"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "-e",
        "--event-list",
        type=str,
        default=None,
        help="File containing event list",
    )
    parser.add_argument(
        "-l",
        "--lc",
        type=str,
        default=None,
        help="File containing light curve",
    )
    parser.add_argument(
        "-c",
        "--ctrate",
        type=float,
        default=None,
        help="Count rate for simulated events",
    )
    parser.add_argument(
        "-o",
        "--outname",
        type=str,
        default="events.evt",
        help="Output file name",
    )
    parser.add_argument(
        "-i", "--instrument", type=str, default=None, help="Instrument name"
    )
    parser.add_argument(
        "-m", "--mission", type=str, default=None, help="Mission name"
    )
    parser.add_argument(
        "--tstart",
        type=float,
        default=None,
        help="Start time of the observation (s from MJDREF)",
    )
    parser.add_argument(
        "--tstop",
        type=float,
        default=None,
        help="End time of the observation (s from MJDREF)",
    )
    parser.add_argument(
        "--mjdref", type=float, default=55197.00076601852, help="Reference MJD"
    )
    parser.add_argument(
        "--deadtime",
        type=float,
        default=None,
        nargs="+",
        help="Dead time magnitude. Can be specified as a "
        "single number, or two. In this last case, the "
        "second value is used as sigma of the dead time "
        "distribution",
    )
    args = check_negative_numbers_in_args(args)
    _add_default_args(parser, ["loglevel", "debug"])

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENfake.log"):
        additional_columns = {}
        livetime = None
        if (
            args.lc is None
            and args.ctrate is None
            and args.event_list is not None
        ):
            event_list = _read_event_list(args.event_list)
        elif args.lc is not None or args.ctrate is not None:
            event_list = EventList()
            if args.lc is not None:
                lc = _read_light_curve(args.lc)
            elif args.ctrate is not None:
                tstart = assign_value_if_none(args.tstart, 0)
                tstop = assign_value_if_none(args.tstop, 1024)
                dt = (tstop - tstart) / 1024
                t = np.arange(tstart, tstop + 1, dt)
                lc = Lightcurve(
                    time=t, counts=args.ctrate + np.zeros_like(t), dt=dt
                )
            event_list.simulate_times(lc)
            nevents = len(event_list.time)
            event_list.pi = np.zeros(nevents, dtype=int)
            event_list.mjdref = args.mjdref
            log.info("{} events generated".format(nevents))
        else:
            event_list = None

        if args.deadtime is not None and event_list is not None:
            deadtime = args.deadtime[0]
            deadtime_sigma = None
            if len(args.deadtime) > 1:
                deadtime_sigma = args.deadtime[1]
            event_list, info = filter_for_deadtime(
                event_list, deadtime, dt_sigma=deadtime_sigma, return_all=True
            )

            log.info("{} events after filter".format(len(event_list.time)))

            prior = np.zeros_like(event_list.time)

            prior[1:] = np.diff(event_list.time) - info.deadtime[:-1]

            additional_columns["PRIOR"] = {"data": prior, "format": "D"}
            additional_columns["KIND"] = {"data": info.is_event, "format": "L"}
            livetime = np.sum(prior)

        generate_fake_fits_observation(
            event_list=event_list,
            filename=args.outname,
            instr=args.instrument,
            mission=args.mission,
            tstart=args.tstart,
            tstop=args.tstop,
            mjdref=args.mjdref,
            livetime=livetime,
            additional_columns=additional_columns,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test a full run of the codes from the command line."""

import os

from astropy.tests.helper import remote_data
import pytest

import hendrics as hen
from hendrics.tests import _dummy_par
from hendrics.fold import HAS_PINT
from hendrics import binary, calibrate, io, lcurve, read_events

HEN_FILE_EXTENSION = hen.io.HEN_FILE_EXTENSION

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


class TestBinary(object):
    """Test how command lines work.

    Usually considered bad practice, but in this
    case I need to test the full run of the codes, and files depend on each
    other.
    Inspired by http://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order

    When command line is missing, uses some function calls
    """  # NOQA

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.par = _dummy_par("bubububu.par")

        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileB = os.path.join(
            cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileAcal = os.path.join(
            cls.datadir,
            "monol_testA_nustar_fpma_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.par = _dummy_par("bubububu.par")
        command = "{0} --discard-calibration".format(
            os.path.join(cls.datadir, "monol_testA.evt")
        )
        hen.read_events.main(command.split())
        command = "{} -r {}".format(
            os.path.join(
                cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(cls.datadir, "test.rmf"),
        )
        hen.calibrate.main(command.split())
        cls.lcA = os.path.join(
            os.path.join(
                cls.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )
        command = (
            "{} -e 3 50 --safe-interval 100 300  --nproc 2 -b 0.5 " "-o {}"
        ).format(cls.ev_fileAcal, cls.lcA)
        hen.lcurve.main(command.split())

    def test_save_binary_events(self):
        f = self.ev_fileA
        with pytest.raises(ValueError) as excinfo:
            hen.binary.main_presto(
                "{} -b 0.1 -e 3 59 --debug".format(f).split()
            )

        assert "Energy filtering requested" in str(excinfo.value)

    @remote_data
    @pytest.mark.skipif("not HAS_PINT")
    def test_save_binary_calibrated_events(self):
        f = self.ev_fileAcal
        hen.binary.main_presto(
            "{} -b 0.1 -e 3 59 --debug --deorbit-par {}".format(
                f, self.par
            ).split()
        )
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, ".dat"))
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, ".inf"))

    def test_save_binary_lc(self):
        f = self.lcA
        hen.binary.main_presto("{}".format(f).split())
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, ".dat"))
        assert os.path.exists(f.replace(HEN_FILE_EXTENSION, ".inf"))



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import shutil
import os
import glob

import numpy as np
from astropy import log
from astropy.logger import AstropyUserWarning
from astropy.tests.helper import remote_data
import pytest
from stingray.lightcurve import Lightcurve
import hendrics as hen
from hendrics.tests import _dummy_par
from hendrics.fold import HAS_PINT
from hendrics import (
    fake,
    fspec,
    base,
    calibrate,
    create_gti,
    exposure,
    exvar,
    io,
    lcurve,
    plot,
    read_events,
    rebin,
)

from hendrics.read_events import treat_event_file
from hendrics.io import HEN_FILE_EXTENSION, get_file_type
from hendrics.lcurve import lcurve_from_events

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError

log.setLevel("DEBUG")


class TestLcurve:
    """Real unit tests."""

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")
        cls.fits_fileA = os.path.join(cls.datadir, "monol_testA.evt")
        cls.new_filename = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.calib_filename = os.path.join(
            cls.datadir,
            "monol_testA_nustar_fpma_ev_calib" + HEN_FILE_EXTENSION,
        )

    def test_treat_event_file_nustar(self):
        from astropy.io.fits import Header

        treat_event_file(self.fits_fileA, discard_calibration=True)
        lcurve_from_events(self.new_filename)
        newfile = os.path.join(
            self.datadir, "monol_testA_nustar_fpma_lc" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(newfile)
        type, data = get_file_type(newfile)
        assert type == "lc"
        assert isinstance(data, Lightcurve)
        Header.fromstring(data.header)
        assert hasattr(data, "mjdref")
        assert data.mjdref > 0

    def test_treat_event_file_nustar_energy(self):
        command = "{0} -r {1} --nproc 2".format(
            self.new_filename, os.path.join(self.datadir, "test.rmf")
        )
        hen.calibrate.main(command.split())
        lcurve_from_events(self.calib_filename, e_interval=[3, 50])

        newfile = os.path.join(
            self.datadir,
            "monol_testA_nustar_fpma_E3-50_lc" + HEN_FILE_EXTENSION,
        )
        assert os.path.exists(newfile)
        type, data = get_file_type(newfile)
        assert type == "lc"
        assert isinstance(data, Lightcurve)
        assert hasattr(data, "mjdref")
        assert data.mjdref > 0


class TestFullRun(object):
    """Test how command lines work.

    Usually considered bad practice, but in this
    case I need to test the full run of the codes, and files depend on each
    other.
    Inspired by http://stackoverflow.com/questions/5387299/python-unittest-testcase-execution-order

    When command line is missing, uses some function calls
    """  # NOQA

    @classmethod
    def setup_class(cls):
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")

        cls.ev_fileA = os.path.join(
            cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileB = os.path.join(
            cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
        )
        cls.ev_fileAcal = os.path.join(
            cls.datadir,
            "monol_testA_nustar_fpma_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.ev_fileBcal = os.path.join(
            cls.datadir,
            "monol_testB_nustar_fpmb_ev_calib" + HEN_FILE_EXTENSION,
        )
        cls.par = _dummy_par("bubububu.par")
        command = "{0} {1} --discard-calibration".format(
            os.path.join(cls.datadir, "monol_testA.evt"),
            os.path.join(cls.datadir, "monol_testB.evt"),
        )
        hen.read_events.main(command.split())
        command = "{} {} -r {}".format(
            os.path.join(
                cls.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(
                cls.datadir, "monol_testB_nustar_fpmb_ev" + HEN_FILE_EXTENSION
            ),
            os.path.join(cls.datadir, "test.rmf"),
        )
        hen.calibrate.main(command.split())

    def test_lcurve(self):
        """Test light curve production."""
        from astropy.io.fits import Header

        new_filename = os.path.join(
            os.path.join(
                self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )
        command = (
            "{0} -e {1} {2} --safe-interval "
            "{3} {4}  --nproc 2 -b 0.5 -o {5}"
        ).format(self.ev_fileAcal, 3, 50, 100, 300, new_filename)
        hen.lcurve.main(command.split())

        assert os.path.exists(new_filename)
        lc = hen.io.load_lcurve(new_filename)
        assert hasattr(lc, "header")
        # Test that the header is correctly conserved
        Header.fromstring(lc.header)
        assert hasattr(lc, "gti")
        gti_to_test = hen.io.load_events(self.ev_fileAcal).gti
        assert np.allclose(gti_to_test, lc.gti)

    def test_lcurve_B(self):
        command = (
            "{0} -e {1} {2} --safe-interval " "{3} {4} -b 0.5 -o {5}"
        ).format(
            self.ev_fileBcal,
            3,
            50,
            100,
            300,
            os.path.join(
                self.datadir, "monol_testB_E3-50_lc" + HEN_FILE_EXTENSION
            ),
        )
        hen.lcurve.main(command.split())
        assert os.path.exists(
            os.path.join(
                self.datadir, "monol_testB_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )

    def test_lcurve_noclobber(self):
        input_file = self.ev_fileAcal
        new_filename = os.path.join(
            os.path.join(
                self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
            )
        )

        with pytest.warns(AstropyUserWarning) as record:
            command = ("{0} -o {1} --noclobber").format(
                input_file, new_filename
            )
            hen.lcurve.main(command.split())
        assert [
            "File exists, and noclobber" in r.message.args[0] for r in record
        ]

    def test_lcurve_split(self):
        """Test lc with gti-split option."""
        command = "{0} {1} -g".format(self.ev_fileAcal, self.ev_fileBcal)
        hen.lcurve.main(command.split())
        new_filename = os.path.join(
            self.datadir,
            "monol_testA_nustar_fpma_gti000_lc" + HEN_FILE_EXTENSION,
        )
        assert os.path.exists(new_filename)
        lc = hen.io.load_lcurve(new_filename)
        gti_to_test = hen.io.load_events(self.ev_fileAcal).gti[0]
        assert np.allclose(gti_to_test, lc.gti)

    def test_fits_lcurve0(self):
        """Test light curves from FITS."""
        lcurve_ftools_orig = os.path.join(self.datadir, "lcurveA.fits")

        lcurve_ftools = os.path.join(
            self.datadir, "lcurve_ftools_lc" + HEN_FILE_EXTENSION
        )

        command = "{0} --outfile {1}".format(
            self.ev_fileAcal, os.path.join(self.datadir, "lcurve_lc")
        )
        hen.lcurve.main(command.split())
        assert os.path.exists(
            os.path.join(self.datadir, "lcurve_lc") + HEN_FILE_EXTENSION
        )

        command = "--fits-input {0} --outfile {1}".format(
            lcurve_ftools_orig, lcurve_ftools
        )
        hen.lcurve.main(command.split())
        with pytest.warns(AstropyUserWarning) as record:
            command = command + " --noclobber"
            hen.lcurve.main(command.split())
        assert [
            "File exists, and noclobber" in r.message.args[0] for r in record
        ]

    def test_fits_lcurve1(self):
        """Test light curves from FITS."""
        lcurve_ftools = os.path.join(
            self.datadir, "lcurve_ftools_lc" + HEN_FILE_EXTENSION
        )

        lcurve_mp = os.path.join(
            self.datadir, "lcurve_lc" + HEN_FILE_EXTENSION
        )

        lcdata_mp = hen.io.load_data(lcurve_mp)
        lcdata_ftools = hen.io.load_data(lcurve_ftools)

        lc_mp = lcdata_mp["counts"]

        lenmp = len(lc_mp)
        lc_ftools = lcdata_ftools["counts"]
        lenftools = len(lc_ftools)
        goodlen = min([lenftools, lenmp])

        diff = lc_mp[:goodlen] - lc_ftools[:goodlen]

        assert np.all(
            np.abs(diff) <= 1e-3
        ), "Light curve data do not coincide between FITS and HEN"

    def test_txt_lcurve(self):
        """Test light curves from txt."""
        lcurve_mp = os.path.join(
            self.datadir, "lcurve_lc" + HEN_FILE_EXTENSION
        )
        lcdata_mp = hen.io.load_data(lcurve_mp)
        lc_mp = lcdata_mp["counts"]
        time_mp = lcdata_mp["time"]

        lcurve_txt_orig = os.path.join(self.datadir, "lcurve_txt_lc.txt")

        hen.io.save_as_ascii([time_mp, lc_mp], lcurve_txt_orig)

        lcurve_txt = os.path.join(
            self.datadir, "lcurve_txt_lc" + HEN_FILE_EXTENSION
        )
        command = "--txt-input " + lcurve_txt_orig + " --outfile " + lcurve_txt
        hen.lcurve.main(command.split())
        lcdata_txt = hen.io.load_data(lcurve_txt)

        lc_txt = lcdata_txt["counts"]

        assert np.all(
            np.abs(lc_mp - lc_txt) <= 1e-3
        ), "Light curve data do not coincide between txt and HEN"

        with pytest.warns(AstropyUserWarning) as record:
            command = command + " --noclobber"
            hen.lcurve.main(command.split())
        assert [
            "File exists, and noclobber" in r.message.args[0] for r in record
        ]

    def test_joinlcs(self):
        """Test produce joined light curves."""
        new_filename = os.path.join(
            self.datadir, "monol_test_joinlc" + HEN_FILE_EXTENSION
        )
        # because join_lightcurves separates by instrument
        new_actual_filename = os.path.join(
            self.datadir, "fpmamonol_test_joinlc" + HEN_FILE_EXTENSION
        )
        lcA_pattern = "monol_testA_nustar_fpma_gti[0-9][0-9][0-9]_lc*"
        lcB_pattern = "monol_testB_nustar_fpmb_gti[0-9][0-9][0-9]_lc*"
        hen.lcurve.join_lightcurves(
            glob.glob(
                os.path.join(self.datadir, lcA_pattern + HEN_FILE_EXTENSION)
            )
            + glob.glob(
                os.path.join(self.datadir, lcB_pattern + HEN_FILE_EXTENSION)
            ),
            new_filename,
        )

        lc = hen.io.load_lcurve(new_actual_filename)
        assert hasattr(lc, "gti")
        gti_to_test = hen.io.load_events(self.ev_fileA).gti
        assert np.allclose(gti_to_test, lc.gti)

    def test_scrunchlcs(self):
        """Test produce scrunched light curves."""
        a_in = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        b_in = os.path.join(
            self.datadir, "monol_testB_E3-50_lc" + HEN_FILE_EXTENSION
        )
        out = os.path.join(
            self.datadir, "monol_test_scrunchlc" + HEN_FILE_EXTENSION
        )
        command = "{0} {1} -o {2}".format(a_in, b_in, out)

        a_lc = hen.io.load_lcurve(a_in)
        b_lc = hen.io.load_lcurve(b_in)
        a_lc.apply_gtis()
        b_lc.apply_gtis()
        hen.lcurve.scrunch_main(command.split())
        out_lc = hen.io.load_lcurve(out)
        out_lc.apply_gtis()
        assert np.all(out_lc.counts == a_lc.counts + b_lc.counts)
        gti_to_test = hen.io.load_events(self.ev_fileA).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def testbaselinelc(self):
        """Test produce scrunched light curves."""
        a_in = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        out = os.path.join(self.datadir, "monol_test_baselc")
        command = "{0} -o {1} -p 0.001 --lam 1e5".format(a_in, out)

        hen.lcurve.baseline_main(command.split())
        out_lc = hen.io.load_lcurve(out + "_0" + HEN_FILE_EXTENSION)
        assert hasattr(out_lc, "base")
        gti_to_test = hen.io.load_events(self.ev_fileA).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def testbaselinelc_nooutroot(self):
        """Test produce scrunched light curves."""
        a_in = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        command = "{0} -p 0.001 --lam 1e5".format(a_in)

        hen.lcurve.baseline_main(command.split())
        out_lc = hen.io.load_lcurve(
            hen.base.hen_root(a_in) + "_lc_baseline" + HEN_FILE_EXTENSION
        )
        assert hasattr(out_lc, "base")
        gti_to_test = hen.io.load_events(self.ev_fileA).gti
        assert np.allclose(gti_to_test, out_lc.gti)

    def test_lcurve_error_uncalibrated(self):
        """Test light curve error from uncalibrated file."""
        command = ("{0} -e {1} {2}").format(
            os.path.join(
                self.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
            ),
            3,
            50,
        )

        with pytest.raises(ValueError) as excinfo:
            hen.lcurve.main(command.split())
        message = str(excinfo.value)
        assert str(message).strip().endswith("Did you run HENcalibrate?")

    def test_lcurve_pi_filtering(self):
        """Test light curve using PI filtering."""
        command = ("{0} --pi-interval {1} {2}").format(
            os.path.join(
                self.datadir, "monol_testA_nustar_fpma_ev" + HEN_FILE_EXTENSION
            ),
            10,
            300,
        )

        hen.lcurve.main(command.split())

    def test_rebinlc(self):
        """Test LC rebinning."""
        command = "{0} -r 4".format(
            os.path.join(self.datadir, "monol_testA_E3-50_lc")
            + HEN_FILE_EXTENSION
        )
        hen.rebin.main(command.split())

    def test_save_fvar_from_lc(self):
        fname = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        hen.exvar.main(
            [fname, "-c", "10", "--fraction-step", "0.6", "--norm", "fvar"]
        )
        out = hen.base.hen_root(fname) + "_fvar" + ".qdp"
        os.path.exists(out)

    def test_save_excvar_from_lc(self):
        fname = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        hen.exvar.main([fname])
        out = hen.base.hen_root(fname) + "_excvar" + ".qdp"
        os.path.exists(out)

    def test_save_excvar_norm_from_lc(self):
        fname = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        hen.exvar.main([fname, "--norm", "norm_excvar"])
        out = hen.base.hen_root(fname) + "_norm_excvar" + ".qdp"
        os.path.exists(out)

    def test_save_excvar_wrong_norm_from_lc(self):
        fname = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        with pytest.raises(ValueError) as excinfo:
            hen.exvar.main([fname, "--norm", "cicciput"])
        assert "Normalization must be fvar, " in str(excinfo.value)

    def test_create_gti_lc(self):
        """Test creating a GTI file."""
        fname = (
            os.path.join(self.datadir, "monol_testA_E3-50_lc")
            + HEN_FILE_EXTENSION
        )
        command = "{0} -f counts>0 -c --debug".format(fname)
        hen.create_gti.main(command.split())

    def test_apply_gti_lc(self):
        """Test applying a GTI file."""
        fname = (
            os.path.join(self.datadir, "monol_testA_E3-50_gti")
            + HEN_FILE_EXTENSION
        )
        lcfname = (
            os.path.join(self.datadir, "monol_testA_E3-50_lc")
            + HEN_FILE_EXTENSION
        )
        lcoutname = (
            os.path.join(self.datadir, "monol_testA_E3-50_lc_gtifilt")
            + HEN_FILE_EXTENSION
        )
        command = "{0} -a {1} --debug".format(lcfname, fname)
        hen.create_gti.main(command.split())
        hen.io.load_lcurve(lcoutname)

    def test_plot_lcurve_baseline(self):
        a_in = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        base_file = (
            hen.base.hen_root(a_in) + "_lc_baseline" + HEN_FILE_EXTENSION
        )
        hen.plot.main([base_file, "--noplot", "-o", "dummy_base.qdp"])
        filedata = np.genfromtxt("dummy_base.qdp")

        assert filedata.shape[1] == 3

    def test_pds_fits(self):
        """Test PDS production with light curves obtained from FITS files."""
        lcurve_ftools = os.path.join(
            self.datadir, "lcurve_ftools_lc" + HEN_FILE_EXTENSION
        )
        command = "{0} --save-all -f 128".format(lcurve_ftools)
        hen.fspec.main(command.split())

    def test_pds_txt(self):
        """Test PDS production with light curves obtained from txt files."""
        lcurve_txt = os.path.join(
            self.datadir, "lcurve_txt_lc" + HEN_FILE_EXTENSION
        )
        command = "{0} --save-all -f 128".format(lcurve_txt)
        hen.fspec.main(command.split())

    def test_exposure(self):
        """Test exposure calculations from unfiltered files."""
        lcname = os.path.join(
            self.datadir, "monol_testA_E3-50_lc" + HEN_FILE_EXTENSION
        )
        ufname = os.path.join(self.datadir, "monol_testA_uf.evt")
        command = "{0} {1}".format(lcname, ufname)

        hen.exposure.main(command.split())
        fname = os.path.join(
            self.datadir, "monol_testA_E3-50_lccorr" + HEN_FILE_EXTENSION
        )
        assert os.path.exists(fname)
        ftype, contents = hen.io.get_file_type(fname)

        assert isinstance(contents, Lightcurve)
        assert hasattr(contents, "expo")

    @classmethod
    def teardown_class(self):
        """Test a full run of the scripts (command lines)."""

        def find_file_pattern_in_dir(pattern, directory):
            return glob.glob(os.path.join(directory, pattern))

        patterns = [
            "*monol_test*" + HEN_FILE_EXTENSION,
            "*lcurve*" + HEN_FILE_EXTENSION,
            "*lcurve*.txt",
            "*.log",
            "*monol_test*.dat",
            "*monol_test*.png",
            "*monol_test*.txt",
            "*monol_test_fake*.evt",
            "*bubu*",
            "*.p",
            "*.qdp",
            "*.inf",
        ]

        file_list = []
        for pattern in patterns:
            file_list.extend(find_file_pattern_in_dir(pattern, self.datadir))

        for f in file_list:
            if os.path.exists(f):
                print("Removing " + f)
                os.remove(f)

        patterns = ["*_pds*/", "*_cpds*/", "*_sum/"]

        dir_list = []
        for pattern in patterns:
            dir_list.extend(find_file_pattern_in_dir(pattern, self.datadir))
        for f in dir_list:
            if os.path.exists(f):
                shutil.rmtree(f)



import os
import copy

import numpy as np
from astropy import log
from stingray.modeling import fit_powerspectrum
from .io import load_model, load_pds, save_model, save_pds, HEN_FILE_EXTENSION


def main_model(args=None):
    """Main function called by the `HENfspec` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Fit frequency spectra (PDS, CPDS, cospectrum) "
        "with user-defined models"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of light curve files", nargs="+")
    parser.add_argument(
        "-m",
        "--modelfile",
        type=str,
        help="File containing an Astropy model with or without" " constraints",
    )
    parser.add_argument(
        "--fitmethod",
        type=str,
        default="L-BFGS-B",
        help="Any scipy-compatible fit method",
    )
    parser.add_argument(
        "--frequency-interval",
        type=float,
        nargs="+",
        default=None,
        help="Select frequency interval(s) to fit. Must be "
        "an even number of frequencies in Hz, like "
        '"--frequency-interval 0 2" or '
        '"--frequency-interval 0 2 5 10", meaning that '
        "the spectrum will be fitted between 0 and 2 Hz, "
        "or using the intervals 0-2 Hz and 5-10 Hz.",
    )

    _add_default_args(parser, ["loglevel", "debug"])

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    freqs = args.frequency_interval
    if freqs is not None and len(freqs) % 2 != 0:
        raise ValueError("Invalid number of frequencies specified")

    log.setLevel(args.loglevel)
    with log.log_to_file("HENmodel.log"):
        model, kind, constraints = load_model(args.modelfile)
        if kind != "Astropy":
            raise TypeError("At the moment, only Astropy models are accepted")

        for f in args.files:
            root = os.path.splitext(f)[0]
            spectrum = load_pds(f)

            if freqs is not None:
                good = np.zeros(len(spectrum.freq), dtype=bool)
                for f0, f1 in zip(freqs[::2], freqs[1::2]):
                    local_good = (spectrum.freq >= f0) & (spectrum.freq < f1)
                    good[local_good] = True

                spectrum_filt = copy.copy(spectrum)
                spectrum_filt.power = spectrum.power[good]
                spectrum_filt.freq = spectrum.freq[good]
                spectrum_filt.power_err = spectrum.power_err[good]

            priors = None
            max_post = False

            if constraints is not None and "priors" in constraints:
                priors = constraints["priors"]
                max_post = True

            parest, res = fit_powerspectrum(
                spectrum,
                model,
                model.parameters,
                max_post=max_post,
                priors=priors,
                fitmethod=args.fitmethod,
            )

            save_model(res.model, root + "_bestfit.p")
            spectrum.best_fits = [res.model]
            log.info("Best-fit model:")
            log.info(res.model)
            save_pds(spectrum, root + "_fit" + HEN_FILE_EXTENSION)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to calculate colors and hardness."""

import os
from astropy import log
from stingray.lightcurve import Lightcurve
import numpy as np
from .io import HEN_FILE_EXTENSION, load_lcurve, save_lcurve
from .base import hen_root
from .lcurve import main as henlcurve


def colors():
    pass


def main(args=None):
    """Main function called by the `HENcolors` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = "Calculate color light curves"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "-e",
        "--energies",
        nargs=4,
        required=True,
        type=str,
        default=None,
        help="The energy boundaries in keV used to calculate "
        "the color. E.g. -e 2 3 4 6 means that the "
        "color will be calculated as 4.-6./2.-3. keV. "
        "If --use-pi is specified, these are interpreted "
        "as PI channels",
    )

    args = check_negative_numbers_in_args(args)
    _add_default_args(
        parser, ["bintime", "usepi", "output", "loglevel", "debug"]
    )
    args = parser.parse_args(args)
    files = args.files
    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)

    with log.log_to_file("HENcolors.log"):
        option = "--energy-interval"
        if args.use_pi:
            option = "--pi-interval"

        if args.outfile is not None and len(files) > 1:
            raise ValueError(
                "Specify --output only when processing " "a single file"
            )
        for f in files:
            henlcurve(
                [f]
                + [option]
                + args.energies[:2]
                + [
                    "-b",
                    str(args.bintime),
                    "-d",
                    ".",
                    "-o",
                    "lc0" + HEN_FILE_EXTENSION,
                ]
            )
            lc0 = load_lcurve("lc0" + HEN_FILE_EXTENSION)
            henlcurve(
                [f]
                + [option]
                + args.energies[2:]
                + [
                    "-b",
                    str(args.bintime),
                    "-d",
                    ".",
                    "-o",
                    "lc1" + HEN_FILE_EXTENSION,
                ]
            )
            lc1 = load_lcurve("lc1" + HEN_FILE_EXTENSION)

            time = lc0.time
            counts = lc1.countrate / lc0.countrate
            counts_err = np.sqrt(
                lc1.countrate_err ** 2 + lc0.countrate_err ** 2
            )
            scolor = Lightcurve(
                time=time,
                counts=counts,
                err=counts_err,
                input_counts=False,
                err_dist="gauss",
                gti=lc0.gti,
                dt=args.bintime,
            )
            del lc0
            del lc1
            os.unlink("lc0" + HEN_FILE_EXTENSION)
            os.unlink("lc1" + HEN_FILE_EXTENSION)

            if args.outfile is None:
                label = "_E_"
                if args.use_pi:
                    label = "_PI_"
                label += "{3}-{2}_over_{1}-{0}".format(*args.energies)
                args.outfile = hen_root(f) + label + HEN_FILE_EXTENSION
            scolor.e_intervals = np.asarray([float(k) for k in args.energies])
            scolor.use_pi = args.use_pi
            save_lcurve(scolor, args.outfile, lctype="Color")
